@@ -1,5 +1,7 @@
 import functools
+from inspect import isclass
 import secrets
+import threading
 import uuid
 import pika
 import pika.exceptions
@@ -10,7 +12,7 @@ from enum import Enum
 from json import JSONEncoder, JSONDecoder
 from threading import Thread, get_ident
 from time import sleep
-from typing import Callable
+from typing import Callable, Type
 from pika.exchange_type import ExchangeType
 
 FAILED_MESSAGES_DLX = '_failed_messages_dlx'
@@ -39,10 +41,11 @@ class Queue:
     name: str
     bindings: list[QueueBinding] = field(default_factory=list)
     passive: bool = False
-    durable: bool = False
+    durable: bool = True
     exclusive: bool = False
     auto_delete: bool = False
     max_priority: int = 0
+    consumer_timeout: int = None
     failure_method: FailureMethod = FailureMethod.DROP
 
 
@@ -61,7 +64,9 @@ class Message:
     data: dict
     delivery_tag: int
     topic: str
+    id: str = None
     exchange: str = None
+    redelivered: bool = False
     first_death: Death = None
     deaths: list[Death] = field(default_factory=list)
 
@@ -72,7 +77,15 @@ class Consumer:
     handler: Callable
     ack_method: AckMethod
     tag: str = ''
-    thread: Thread | None = None
+    threads: list[Thread] = field(default_factory=list)
+    processed_messages: int = 0
+    failed_messages: int = 0
+
+    def join(self):
+        threads = self.threads.copy()
+        for thread in threads:
+            if thread.is_alive():
+                thread.join()
 
 
 @dataclass
@@ -83,10 +96,18 @@ class Heartbeat:
     consumer: Consumer | None = None
 
 
+class Midlleware:
+    def __init__(self, app: 'Snowshoe'):
+        self._app = app
+
+    def __call__(self, message: 'Message', next: Callable):
+        return next(message)
+
+
 class Snowshoe:
-    connection: pika.BlockingConnection
-    consumer_channel: pika.adapters.blocking_connection.BlockingChannel
-    producer_channel: pika.adapters.blocking_connection.BlockingChannel
+    connection: pika.BlockingConnection | None
+    consumer_channel: pika.adapters.blocking_connection.BlockingChannel | None
+    producer_channel: pika.adapters.blocking_connection.BlockingChannel | None
     name: str
 
     def __init__(
@@ -101,9 +122,9 @@ class Snowshoe:
             json_encoder_class=JSONEncoder,
             json_decoder_class=JSONDecoder,
     ) -> None:
-        self.connection: pika.BlockingConnection = None
-        self.consumer_channel: pika.adapters.blocking_connection.BlockingChannel = None
-        self.producer_channel: pika.adapters.blocking_connection.BlockingChannel = None
+        self.connection = None
+        self.consumer_channel = None
+        self.producer_channel = None
         self._connection_thread_ident: int = 0
         self._delivery_tags = set()
         self._thread_idents = set()
@@ -118,24 +139,42 @@ class Snowshoe:
 
         self._queues: dict[str, Queue] = {}
         self._consumers: list[Consumer] = []
-        self._heartbeat: Heartbeat = Heartbeat(topic='heartbeat:' + secrets.token_urlsafe(15))
+        self._heartbeat: Heartbeat = Heartbeat(topic='_heartbeat:' + secrets.token_urlsafe(15))
         self._is_healthy = True
-        self.status = 'stopped'
+        self._state = 'stopped'
         self.json_encoder = json_encoder_class()
         self.json_decoder = json_decoder_class()
 
         self.ensure_connections()
         self.producer_channel.exchange_declare(self.name, ExchangeType.topic)
+        self._middlewares: list[Callable] = []
 
     @property
     def is_healthy(self):
         return self._is_healthy
 
-    def ensure_connections(self):
-        while self.status == 'connecting':
+    @property
+    def status(self):
+        return {
+            'is_healthy': self.is_healthy,
+            'state': self._state,
+            'consumers': [
+                {
+                    'handler': consumer.handler.__name__,
+                    'threads': len(consumer.threads),
+                    'processed_messages': consumer.processed_messages,
+                    'failed_messages': consumer.failed_messages
+                }
+                for consumer in self._consumers
+                if consumer != self._heartbeat.consumer
+            ]
+        }
+
+    def ensure_connections(self, redefine_queues: bool = True):
+        while self._state == 'connecting':
             sleep(1)
-        self.status = 'connecting'
-        if not self.connection or self.connection.is_closed:
+        self._state = 'connecting'
+        if not self.connection or self.connection.is_closed or self.consumer_channel.is_closed or self.producer_channel.is_closed:
             try:
                 self.connection = pika.BlockingConnection(self._connection_parameters)
                 self.consumer_channel = self.connection.channel()
@@ -143,18 +182,29 @@ class Snowshoe:
                 self._delivery_tags.clear()
                 self._thread_idents.clear()
                 self._connection_thread_ident = get_ident()
-                self.status = 'running'
+                self._state = 'running'
             except Exception as e:
-                self.status = 'stopped'
+                self._state = 'stopped'
                 raise e
             self.consumer_channel.basic_qos(prefetch_count=self.concurrency)
             for consumer in self._consumers:
                 consumer.tag = None
 
-            self.define_queues(list(self._queues.values()))
+            if redefine_queues:
+                self.define_queues(list(self._queues.values()))
             self.resume()
-        self.status = 'running'
+        self._state = 'running'
 
+    def _apply_middlewares(self, handler):
+        def build_chain(index):
+            if index < len(self._middlewares):
+                def next_middleware(req):
+                    return self._middlewares[index](req, build_chain(index + 1))
+                return next_middleware
+            else:
+                return handler
+        return build_chain(0)
+    
     def _cancel(self, consumer_tag: str):
         return self.consumer_channel.basic_cancel(consumer_tag)
 
@@ -165,8 +215,9 @@ class Snowshoe:
             body=self.json_encoder.encode(data).encode(),
             properties=pika.BasicProperties(
                 content_type='application/json',
-                expiration=ttl,
-                priority=priority
+                expiration=str(ttl) if ttl else None,
+                priority=priority,
+                headers={'x-message-id': str(uuid.uuid4())}
             )
         )
 
@@ -195,7 +246,7 @@ class Snowshoe:
                 self._heartbeat.received = message.data['sequence']
 
         def mouth():
-            while self.status == 'running':
+            while self._state == 'running':
                 sleep(10)
                 if self.connection.is_closed:
                     self._is_healthy = False
@@ -204,18 +255,17 @@ class Snowshoe:
                 self._heartbeat.sent += 1
                 self.emit(self._heartbeat.topic, {'sequence': self._heartbeat.sent})
 
-
         queue = Queue(
             name='heartbeat[' + str(uuid.uuid4()) + ']',
             bindings=[QueueBinding(self.name, self._heartbeat.topic)],
             exclusive=True,
-            auto_delete=True
+            auto_delete=True,
+            durable=False,
         )
         self.define_queues([queue])
         self._heartbeat.consumer = self.on(queue, AckMethod.INSTANTLY)(ear)
 
-        # self.emit('_heartbeat', {'sequence': self._heartbeat.sent})
-        Thread(target=mouth).start()
+        Thread(target=mouth, daemon=True).start()
 
     def _run(self):
         while True:
@@ -232,15 +282,23 @@ class Snowshoe:
                 self.consumer_channel.stop_consuming()
                 break
 
-
         for consumer in self._consumers:
             if consumer != self._heartbeat.consumer:
-                consumer.thread.join()
+                consumer.join()
 
         self.connection.close()
 
+    def use(self, middleware: Callable | Type[Midlleware], *args, **kwargs):
+        if isclass(middleware) and issubclass(middleware, Midlleware):
+            middleware = middleware(self, *args, **kwargs)
+
+        if not callable(middleware):
+            raise Exception('Middleware must be callable or a Midlleware class')
+        
+        self._middlewares.append(middleware)
+
     def run(self, wait: bool = True):
-        self.status = 'running'
+        self._state = 'running'
         self._echo()
 
         thread = Thread(target=self._run, daemon=True)
@@ -257,8 +315,8 @@ class Snowshoe:
                 consumer.tag = None
 
         for consumer in self._consumers:
-            if consumer != self._heartbeat.consumer and consumer.thread and consumer.thread.is_alive():
-                consumer.thread.join()
+            if consumer != self._heartbeat.consumer:
+                consumer.join()
 
     def resume(self):
         not_running_consumers = [consumer for consumer in self._consumers if consumer.tag is None]
@@ -277,6 +335,8 @@ class Snowshoe:
                 arguments['x-dead-letter-exchange'] = FAILED_MESSAGES_DLX
             if queue.max_priority:
                 arguments['x-max-priority'] = queue.max_priority
+            if queue.consumer_timeout is not None:
+                arguments['x-consumer-timeout'] = queue.consumer_timeout
 
             try:
                 self.consumer_channel.queue_declare(
@@ -287,8 +347,8 @@ class Snowshoe:
                     auto_delete=queue.auto_delete,
                     arguments=arguments
                 )
-            except pika.exceptions.ChannelClosedByBroker as e:
-                self.ensure_connections()
+            except pika.exceptions.ChannelClosedByBroker:
+                self.ensure_connections(redefine_queues=False)
                 self.consumer_channel.queue_delete(queue_name, if_unused=not force, if_empty=not force)
                 self.consumer_channel.queue_declare(
                     queue_name,
@@ -309,7 +369,6 @@ class Snowshoe:
         else:
             self.connection.add_callback_threadsafe(callback)
 
-
     def emit(self, topic: str, data: dict, ttl: int = None, priority: int = None):
         self.ensure_connections()
         self._call_threadsafe(functools.partial(self._emit, topic=topic, data=data, ttl=ttl, priority=priority))
@@ -333,6 +392,7 @@ class Snowshoe:
             raise Exception('Invalid Queue')
 
         def wrapper(handler: Callable[[Message], any]) -> Consumer:
+            handler = self._apply_middlewares(handler=handler)
 
             def do_works(
                     _channel: pika.adapters.blocking_connection.BlockingChannel,
@@ -340,51 +400,57 @@ class Snowshoe:
                     properties: pika.spec.BasicProperties,
                     body: bytes
             ):
-                try:
-                    if properties.headers and properties.headers.get('x-first-death-exchange'):
-                        deaths = [
-                            Death(
-                                count=item['count'],
-                                reason=item['reason'],
-                                queue=item['queue'],
-                                time=item['time'],
-                                exchange=item['exchange'],
-                                routing_keys=item['routing-keys']
+                if properties.headers and properties.headers.get('x-first-death-exchange'):
+                    deaths = [
+                        Death(
+                            count=item['count'],
+                            reason=item['reason'],
+                            queue=item['queue'],
+                            time=item['time'],
+                            exchange=item['exchange'],
+                            routing_keys=item['routing-keys']
+                        )
+                        for item in properties.headers.get('x-death', [])
+                    ]
+                    first_death = next((
+                        death
+                        for death in deaths
+                        if (
+                            death.queue == properties.headers['x-first-death-queue']
+                            and death.reason == properties.headers['x-first-death-reason']
+                            and death.exchange == properties.headers['x-first-death-exchange']
                             )
-                            for item in properties.headers.get('x-death', [])
-                        ]
-                        first_death = next((
-                            death
-                            for death in deaths
-                            if (
-                                death.queue == properties.headers['x-first-death-queue']
-                                and death.reason == properties.headers['x-first-death-reason']
-                                and death.exchange == properties.headers['x-first-death-exchange']
-                                )
-                        ), None)
-                    else:
-                        deaths = []
-                        first_death = None
+                    ), None)
+                else:
+                    deaths = []
+                    first_death = None
 
-                    if ack_method != AckMethod.INSTANTLY:
-                        self._delivery_tags.add(method.delivery_tag)
+                if ack_method != AckMethod.INSTANTLY:
+                    self._delivery_tags.add(method.delivery_tag)
 
-                    message = Message(
-                        data=self.json_decoder.decode(body.decode()),
-                        topic=method.routing_key,
-                        delivery_tag=method.delivery_tag,
-                        exchange=method.exchange,
-                        first_death=first_death,
-                        deaths=deaths
-                    )
+                message = Message(
+                    id=properties.headers and properties.headers.get('x-message-id'),
+                    data=self.json_decoder.decode(body.decode()),
+                    topic=method.routing_key,
+                    delivery_tag=method.delivery_tag,
+                    exchange=method.exchange,
+                    redelivered=method.redelivered,
+                    first_death=first_death,
+                    deaths=deaths
+                )
+                try:
                     result = handler(message)
                     if ack_method == AckMethod.AUTO:
                         self.ack(method.delivery_tag)
+                    consumer.processed_messages += 1
                     return result
                 except Exception as e:
+                    consumer.failed_messages += 1
                     if ack_method == AckMethod.AUTO:
                         self.nack(method.delivery_tag, requeue=queue.failure_method == FailureMethod.REQUEUE)
                     raise e
+                finally:
+                    consumer.threads.remove(threading.current_thread())
 
             consumer = Consumer(
                 queue=queue,
@@ -396,7 +462,7 @@ class Snowshoe:
 
             def callback(*args, **kwargs):
                 thread = Thread(target=do_works, args=args, kwargs=kwargs)
-                consumer.thread = thread
+                consumer.threads.append(thread)
                 thread.start()
                 return thread
 
